use std::error::Error;
use std::path::PathBuf;

use super::position::Span;


/// Represents an error encountered during lexical analysis.
///
/// `LexerError` contains contextual information about where the error
/// occurred in the source input, an optional offending snippet,
/// the file path (if applicable), and an optional underlying error
/// that triggered the failure.
///
/// It is designed to provide detailed diagnostics for file-based or
/// in-memory parsing and is compatible with error reporting ecosystems.
#[derive(Debug)]
pub struct LexerError
{
   /// A human-readable error message.
   pub message: String,

   /// The span where the error occurred.
   pub span: Span,

   /// The file that the error occurred in, if known.
   pub file: Option<PathBuf>,

   /// The source snippet related to the error, if known.
   pub snippet: Option<String>,

   /// An optional underlying error that caused this one.
   pub source: Option<Box<dyn Error>>
}

impl LexerError
{
   /// Creates a new `LexerError` with a message, span, and optional context.
   ///
   /// # Parameters
   /// - `message`: A human-readable explanation of the error.
   /// - `span`: The region in the source where the error occurred.
   /// - `file`: An optional path to the file in which the error occurred.
   /// - `snippet`: An optional problematic input string.
   ///
   /// # Returns
   /// A new instance of `LexerError`.
   pub fn new<S, T>(message: S, span: Span, file: Option<T>,
                    snippet: Option<S>)
                    -> Self
      where S: Into<String>,
            T: Into<PathBuf>
   {
      LexerError { message: message.into(),
                   span,
                   file: file.map(Into::into),
                   snippet: snippet.map(Into::into),
                   source: None }
   }

   /// Creates a `LexerError` from only a message and span.
   ///
   /// This is useful when file or snippet context is not available.
   pub fn from_message<S>(message: S, span: Span) -> Self
      where S: Into<String>
   {
      Self::new(message, span, None::<PathBuf>, None::<S>)
   }

   /// Attaches a snippet of the offending source code.
   ///
   /// This is helpful for diagnostics and tooling output.
   pub fn with_snippet<S>(mut self, snippet: S) -> Self
      where S: Into<String>
   {
      self.snippet = Some(snippet.into());
      self
   }

   /// Attaches the path of the file in which the error occurred.
   pub fn with_file<T>(mut self, file: T) -> Self
      where T: Into<PathBuf>
   {
      self.file = Some(file.into());
      self
   }

   /// Wraps a source error that caused this `LexerError`.
   ///
   /// This allows you to chain errors for more detailed diagnostics.
   pub fn with_source<E>(mut self, err: E) -> Self
      where E: Error + 'static
   {
      self.source = Some(Box::new(err));
      self
   }
}

impl std::fmt::Display for LexerError
{
   fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result
   {
      write!(f, "Lexer error at {}", self.span)?;

      if let Some(file) = &self.file
      {
         write!(f, " in file `{}`", file.display())?;
      }

      write!(f, ": {}", self.message)?;

      if let Some(snippet) = &self.snippet
      {
         write!(f, "\n  --> Snippet: `{}`", snippet)?;
      }

      Ok(())
   }
}

impl Error for LexerError
{
   /// Returns the underlying cause of this error, if any.
   fn source(&self) -> Option<&(dyn Error + 'static)>
   {
      self.source.as_ref().map(|e| e.as_ref())
   }
}
