/// Represents a location in a file by line and column.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct Position
{
   /// The referenced line of the file.
   pub line: usize,

   /// The referenced column of the file.
   pub column: usize
}


/// Represents a range within a file.
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub struct Span
{
   /// The start `Position` in the file.
   pub start: Position,

   /// The end `Position` in the file.
   pub end: Position
}



impl Position
{
   /// Creates a new `Position` with the given line and column.
   ///
   /// # Arguments
   /// * `line` - The line number in the source (0-based).
   /// * `column` - The column number within the line (0-based).
   ///
   /// # Returns
   /// A `Position` representing the specified line and column.
   pub fn new(line: usize, column: usize) -> Self
   {
      Position { line, column }
   }
}

impl Default for Position
{
   fn default() -> Self
   {
      Position { line: 0, column: 0 }
   }
}

impl std::fmt::Display for Position
{
   fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result
   {
      write!(f, "Line {}, Column {}", self.line, self.column)
   }
}



impl Span
{
   /// Creates a new `Span` from a starting and ending `Position`.
   ///
   /// # Arguments
   /// * `start` - The starting position of the span.
   /// * `end` - The ending position of the span.
   ///
   /// # Returns
   /// A `Span` covering the range from `start` to `end`.
   pub fn new(start: Position, end: Position) -> Self
   {
      Span { start, end }
   }

   /// Creates a `Span` that covers a single position.
   ///
   /// Useful for zero-length spans or pinpointing a specific token or
   /// character.
   ///
   /// # Arguments
   /// * `val` - The position to be used as both start and end.
   ///
   /// # Returns
   /// A `Span` that starts and ends at `val`.
   pub fn with_single(val: Position) -> Self
   {
      Span { start: val,
             end: val }
   }

   /// Merges two spans into one, taking the start of the first
   /// and the end of the second.
   ///
   /// # Arguments
   /// * `a` - The first span.
   /// * `b` - The second span.
   ///
   /// # Returns
   /// A `Span` that starts at `a.start` and ends at `b.end`.
   ///
   /// # Note
   /// Assumes that `a` comes before `b` in source order.
   pub fn merge(a: Span, b: Span) -> Self
   {
      Span { start: a.start,
             end: b.end }
   }

   /// Merges this span with another, producing a new span
   /// from `self.start` to `other.end`.
   ///
   /// # Arguments
   /// * `other` - Another span to merge with.
   ///
   /// # Returns
   /// A new `Span` from the start of `self` to the end of `other`.
   pub fn merge_with(&self, other: Span) -> Span
   {
      Span { start: self.start,
             end: other.end }
   }
}

impl Default for Span
{
   fn default() -> Self
   {
      Span { start: Position::default(),
             end: Position::default() }
   }
}

impl std::fmt::Display for Span
{
   fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result
   {
      write!(f, "Span[{} -> {}]", self.start, self.end)
   }
}
